# -*- coding: utf-8 -*-
# ========================================================================
#  2023/7/7 17:53
#                 _____   _   _   _____   __   _   _____  
#                /  ___| | | | | | ____| |  \ | | /  ___| 
#                | |     | |_| | | |__   |   \| | | |     
#                | |     |  _  | |  __|  | |\   | | |  _  
#                | |___  | | | | | |___  | | \  | | |_| | 
#                \_____| |_| |_| |_____| |_|  \_| \_____/ 
# ------------------------------------------------------------------------
# 
# 
# 
# ========================================================================

import os
import time
import warnings
import configparser

warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import numpy as np
from pymatgen.core import Structure, Lattice

from predict_structure import PredictStructure
from utils.file_utils import check_and_rename_path, check_path
from utils.read_input import ReadInput
from utils.compound_utils import get_elements_info, get_element_combinations_2, get_element_count_combinations_2
# from utils.children_cell_utils import get_all_children_cell, get_all_children_wyckoff_combination
from utils.children_cell_utils import get_all_children_cell_2, get_children_cell_2
from utils.parameter_utils import parameter_config
from utils.print_utils import print_header, print_run_info
from utils.gen_crystal_WyckPos import GenCrystalWyckPos, get_wyckoffs_strings, wp_alphabet
from utils.gen_crystal_WyckPos_old import get_all_wyckoff_combination
from calculators.m3gnet.origin_m3gnet import OrinM3GNET

class StructureSampling(PredictStructure):

    def create_save_results_data_file(self):
        with open(self.step_result_file, 'w+') as f:
            f.write("number,step,compound,target,sg_number,child_cell,time\n")

    def calculator_config(self):
        super().calculator_config()

        use_relax = False

        calcu1 = OrinM3GNET(
            model_path='path_to_MLIP_model_1',
            use_relax=use_relax)
        calcu2 = OrinM3GNET(
            model_path='path_to_MLIP_model_2',
            use_relax=use_relax)
        calcu3 = OrinM3GNET(
            model_path='path_to_MLIP_model_3',
            use_relax=use_relax)
        self.my_calculators = [calcu1, calcu2, calcu3]

    def get_structure_properties(self, struc_param: dict, **kwargs):
        self.step_number += 1

        try:
            struc, child_cell = self.convert_structure_from_struc_param(struc_param)

            # ############### relax structure traj ###############
            if self.use_relax:
                # print(self.calculator)
                traj_structs, traj_energies = self.calculator.get_relax_traj(struc)
                self.ACS_limit(traj_structs[0])
                mid_i = int(len(traj_energies)/2)
                formula = struc.formula.replace(' ', '')
                st_indies = [0, mid_i, len(traj_energies)-1]
                for i in st_indies:
                    te = traj_energies[i]
                    structure_file_name = os.path.join(self.structures_path,
                                                       '%s_%d_%.5f.cif' % (formula, i, te))
                    traj_structs[i].to(fmt='cif', filename=structure_file_name)

            energies = []

            for calculator in self.my_calculators:
                target_property = calculator.get_target_from_predict(struc)
                energies.append(target_property)

            standard_deviation = np.std(energies)
            error = -standard_deviation
            self.structure_number += 1
            self.save_data_for_successful_step(-error, struc_param, struc, child_cell)

        except Exception as e:
            print(e)
            print(e.__traceback__.tb_frame.f_globals["__file__"])
            print(e.__traceback__.tb_lineno)
            error = 999

        return error

if __name__ == '__main__':
    config = configparser.ConfigParser()
    config.add_section('BASE')
    config.set('BASE', 'atom_element', '[3, 4, 5, 6, 7] [3, 4, 5, 6, 7] [3, 4, 5, 6, 7]')
    config.set('BASE', 'atom_count', '[1-5] [1-5] [1-5]')
    config.set('BASE', 'use_children_cell', 'True')
    config.set('BASE', 'min_atomic_dist_limit', '-0.75')
    config.set('BASE', 'volume_limit', '[0, 0]')
    config.set('BASE', 'max_vacuum_limit', '5.0')
    config.set('BASE', 'output_path', 'path_to_the_output_structures')

    config.add_section('CALCULATOR')
    config.set('CALCULATOR', 'calculator', 'm3gnet')
    config.set('CALCULATOR', 'calculator_path', './MP-2021.2.8-EFS')
    config.set('CALCULATOR', 'use_calculator_relax', 'True')
    config.set('CALCULATOR', 'use_keep_symmetry', 'False')
    config.set('CALCULATOR', 'symprec', '0.001')
    config.set('CALCULATOR', 'use_gpu', 'False')

    config.add_section('OPTIMIZER')
    config.set('OPTIMIZER', 'algorithm', "tpe")
    config.set('OPTIMIZER', 'n_init', '200')
    config.set('OPTIMIZER', 'max_step', '5000')
    config.set('OPTIMIZER', 'rand_seed', '-1')
    config.set('OPTIMIZER', 'use_resume', 'False')
    config.set('OPTIMIZER', 'n_mpi', '1')
    config.set('OPTIMIZER', 'storage', '')

    config.add_section('LATTICE')
    config.set('LATTICE', 'space_group', '[%d-%d]' % (1, 230))
    config.set('LATTICE', 'use_flexible_site', 'False')
    config.set('LATTICE', 'wyck_pos_gen', '4')
    config.set('LATTICE', 'max_wyck_pos_count', '200000')
    config.set('LATTICE', 'lattice_a', '[%f-%f]' % (2.0, 15.0))
    config.set('LATTICE', 'lattice_b', '[%f-%f]' % (2.0, 15.0))
    config.set('LATTICE', 'lattice_c', '[%f-%f]' % (2.0, 15.0))
    config.set('LATTICE', 'lattice_alpha', '[%f-%f]' % (80.0, 100.0))
    config.set('LATTICE', 'lattice_beta', '[%f-%f]' % (80.0, 100.0))
    config.set('LATTICE', 'lattice_gamma', '[%f-%f]' % (80.0, 100.0))

    StructureSampling(None, config)
